#include <stdio.h>
#include <io.h>
#include <dos.h>
#include <sys\stat.h>
#include <sys\types.h>
#include <malloc.h>
#include <fcntl.h>

// 

char magic_bytes_end[] = "Elib";
char magic_bytes_list[] = "Flst";

short int             arc_handle=-1;
struct filelist_entry* filelist=NULL;
struct filelist_entry* cmdfiles=NULL;

// 

struct fileendheader_
{
       long filelistheader;
       char magic_bytes[4];
} fileendheader;

struct filelistheader_
{
       char magic_bytes[4];
       long filelist;
       long filenames;
       short int nr_of_files;
       long len_of_exec;
       short len_of_strings;
} filelistheader;

struct filedate
{
       char year;
       char month;
       char day;
};

struct filetime
{
       char hour;
       char minute;
       char second;
};

struct filelist_entry
{
       long filename;
       long filepos;
       long len;
       struct filedate date;
       struct filetime time;
       long curpos;
};
// 

void free_dir()
{
    free((char*) (filelist[0].filename));
    free(filelist);
    close(arc_handle);
}
// 

char read_dir (char* filename)
{
    long dir_size;
    short i;
    char* pointer;

    arc_handle = open(filename, O_RDONLY | O_BINARY);
    if (arc_handle == -1) return(0);

    lseek( arc_handle, -(long) sizeof( struct fileendheader_ ), SEEK_END );
    read( arc_handle, &fileendheader, sizeof( struct fileendheader_ ) );

    if ( memcmp(fileendheader.magic_bytes, magic_bytes_end, sizeof(magic_bytes_end)-1) )
    {
       return(0);
    }

    lseek( arc_handle, fileendheader.filelistheader, SEEK_SET );
    read( arc_handle, &filelistheader, sizeof( struct filelistheader_ ) );

    if ( memcmp(filelistheader.magic_bytes, magic_bytes_list, sizeof(magic_bytes_list)-1) )
    {
       return(0);
    }

    dir_size = sizeof (struct filelist_entry) * filelistheader.nr_of_files;
    filelist = (struct filelist_entry*) malloc (dir_size);
    if (filelist == NULL) return(0);

    lseek( arc_handle, filelistheader.filelist, SEEK_SET);
    read( arc_handle, filelist, dir_size);

    filelist[0].filename = (long) malloc ( filelistheader.len_of_strings );
    if (filelist[0].filename == 0) return(0);

    lseek( arc_handle, filelistheader.filenames, SEEK_SET);
    read( arc_handle, (char*) (filelist[0].filename), filelistheader.len_of_strings);

    pointer = (char*) (filelist[0].filename);
    for (i=1; i < filelistheader.nr_of_files; i++)
    {
        pointer += ((char*) (filelist[i-1].filename))[0];
        filelist[i].filename = (long) pointer;
    }

    atexit(free_dir);
    return(1);
}
// 

#define MY_HANDLES ('S'*256+'P')

int xopen( char *path, int access, ... )
{
    int i;
    for (i=0;i<filelistheader.nr_of_files;i++)
    {
        if (strcmp(strupr((char*) (filelist[i].filename+1)), strupr(path)) == 0)
        {
           filelist[i].curpos = 0;
           return(MY_HANDLES+i);
        }
    }
    return(open( path, access ));
}

#define IS_HANDLE (handle >= MY_HANDLES && handle <= MY_HANDLES + filelistheader.nr_of_files)

long int xtell( int handle)
{
    if (IS_HANDLE)
       return(filelist[handle-MY_HANDLES].curpos);
    return(tell(handle));
}

long int xlseek( int handle, long int offset, int whence )
{
    if (IS_HANDLE)
    {
       switch (whence)
       {
          case SEEK_SET: filelist[handle-MY_HANDLES].curpos = offset; break;
          case SEEK_CUR: filelist[handle-MY_HANDLES].curpos += offset; break;
          case SEEK_END: filelist[handle-MY_HANDLES].curpos =
                         filelist[handle-MY_HANDLES].len+ offset; break;
       }
       return(0);
    }
    return( lseek( handle, offset, whence ) );
}

int xread( int handle, void *buffer, unsigned len )
{
    int i;

    if (IS_HANDLE)
    {
       i = handle - MY_HANDLES;
       lseek( arc_handle, filelist[i].curpos+ filelist[i].filepos, SEEK_SET);
       if (filelist[i].curpos > filelist[i].len) return(0);
       if (filelist[i].curpos + len > filelist[i].len)
       {
          len = filelist[i].len - filelist[i].curpos;
       }
       read (arc_handle, buffer, len);
       filelist[i].curpos += len;
       return (len);
    }
    return( read( handle, buffer, len ) );
}

int xwrite( int handle, void *buffer, unsigned len )
{
    if (IS_HANDLE) return(0); // This file is read-only
    else
       return( write( handle, buffer, len ) );
}

int xeof( int handle )
{
    if (IS_HANDLE)
       if (filelist[handle-MY_HANDLES].curpos >= filelist[handle-MY_HANDLES].len)
          return(1);
       else return(0);
    return( eof (handle) );
}

int xclose( int handle )
{
    if (!IS_HANDLE) return(close(handle));
    return(0);
}

long int xfilelength( int handle )
{
    if (IS_HANDLE) return(filelist[handle-MY_HANDLES].len);
    return(filelength(handle));
}
